USING: accessors arrays assocs classes grouping hash-sets kernel lists
math math.combinatorics math.order prettyprint sequences
sequences.deep sorting strings vectors ;
IN: texasholdem

: char>string ( char -- str ) 1 swap <string> ;

: 2char>string ( a b -- string ) 2array >string ;


TUPLE: card rank suit ;

: <card> ( cardstr -- card ) [ first ] [ second ] bi card boa ;

: <cards> ( seq -- seq ) [ <card> ] map ;

: worst-hand ( -- seq ) { "2C" "4D" "5D" "6C" "7H" } <cards> ;

: print-card ( card -- ) [ rank>> ] [ suit>> ] bi 2char>string . ;

: =-rank ( card card -- ? ) rank>> swap rank>> = ;

: =-suit ( card card -- ? ) suit>> swap suit>> = ;

: rank>value ( card -- a ) rank>> char>string H{ { "2" 2 } { "3" 3 } { "4" 4 } { "5" 5 } { "6" 6 } { "7" 7 } { "8" 8 } { "9" 9 } { "T" 10 } { "J" 11 } { "Q" 12 } { "K" 13 } { "A" 14 } } at ;

M: card <=> [ rank>value ] bi@ <=> ;

: deck ( -- deck ) "23456789TJQKA" "SCDH" [ 2char>string ] cartesian-map flatten [ <card> ] map ;

: group-by-rank ( hand -- assoc ) H{ } clone swap [ dupd rank>value swap [ dup [ 1 + ] [ drop 1 ] if ] change-at ] each ;

: cmp-length-rank ( x y -- <=> ) 2dup [ second ] bi@ >=< dup +eq+ = [ drop [ first ] bi@ >=< ] [ 2nip ] if ;

: rank-ordered ( hand -- seq ) group-by-rank >alist [ cmp-length-rank ] sort-with ; 

: rank-ordered-rank ( hand -- seq ) rank-ordered [ first ] map ; 

: rank-ordered-num ( hand -- seq ) rank-ordered [ second ] map ;


: n-of-kind? ( hand n -- ? ) <combinations> [ [ =-rank ] monotonic? ] any? ;

: four-of-kind? ( hand -- ? ) 4 n-of-kind? ;

: three-of-kind? ( hand -- ? ) 3 n-of-kind? ;

: pair? ( hand -- ? ) 2 n-of-kind? ;

: full-house? ( hand -- ? ) rank-ordered-num [ first 3 = ] [ second 2 = ] bi and ;

: flush? ( hand -- ? ) [ =-suit ] monotonic? ;

: straight? ( hand -- ? ) sort [ rank>value swap rank>value swap - -1 = ] monotonic? ;

: straight-flush? ( hand -- ? ) [ straight? ] [ flush? ] bi and ;

: two-pair? ( hand -- ? ) <permutations> [ [ [ first ] [ second ] bi =-rank ] [ [ third ] [ fourth ] bi =-rank ] bi and ] any? ;



: keep-not-eq ( <=> <=> -- <=> ) swap dup +eq+ = not [ nip ] [ drop ] if ;

: cmp-same-hands ( hand hand -- <=> ) [ rank-ordered-rank ] bi@ [ <=> ] [ keep-not-eq ] 2map-reduce ;

: cmp-step ( hand hand ? ? quote -- <=> ) -rot 2dup and [ 3drop cmp-same-hands ] [ 2dup [ not ] bi@ and [ 2drop call( hand hand -- <=> ) ] [ [ 4drop ] dip [ +lt+ ] [ +gt+ ] if ] if ] if ;

: cmp-pair ( hand hand -- <=> ) 2dup [ pair? ] bi@ [ cmp-same-hands ] cmp-step ;

: cmp-two-pair ( hand hand -- <=> ) 2dup [ two-pair? ] bi@ [ cmp-pair ] cmp-step ;

: cmp-three-kind ( hand hand -- <=> ) 2dup [ three-of-kind? ] bi@ [ cmp-two-pair ] cmp-step ;

: cmp-straight ( hand hand -- <=> ) 2dup [ straight? ] bi@ [ cmp-three-kind ] cmp-step ;

: cmp-flush ( hand hand -- <=> ) 2dup [ flush? ] bi@ [ cmp-straight ] cmp-step ;

: cmp-full-house ( hand hand -- <=> ) 2dup [ full-house? ] bi@ [ cmp-flush ] cmp-step ;

: cmp-four-kind ( hand hand -- <=> ) 2dup [ four-of-kind? ] bi@ [ cmp-full-house ] cmp-step ;

: cmp-hands ( hand hand -- <=> ) 2dup [ straight-flush? ] bi@ [ cmp-four-kind ] cmp-step ;

: keep-best-hand ( hand hand -- hand ) 2dup cmp-hands +gt+ = [ drop ] [ nip ] if ;


: num-wins ( -- assoc ) H{ } ;

: clear-num-wins ( -- ) num-wins keys [ num-wins delete-at ] each ;

: best-com-hand ( com hand -- hand ) append 5 <combinations> dup first [ keep-best-hand ] reduce ;

: remove-seq ( seq seq -- seq ) [ swap remove ] each ;

: remove-used-cards ( com hands discard deck -- com hands deck ) swap remove-seq [ 2dup ] dip swap [ remove-seq ] each swap remove-seq ;

: ident ( -- ident ) "tie" worst-hand 2array ;

: add-wins ( hands besthands -- ) [ 2array ] 2map ident [ 2dup [ second ] bi@ cmp-hands dup +eq+ = [ 3drop ident ] [ +gt+ = [ drop ] [ nip ] if ] if ] reduce first num-wins [ dup [ 1 + ] [ drop 1 ] if ] change-at ;

: eval-hands ( com hands -- ) tuck [ over best-com-hand ] map nip add-wins ;

: extend ( com hands deck -- ) rot dup length 5 < [ over [ [ 3dup ] dip dup [ 1array append ] dip rot remove swapd extend ] each 3drop ] [ rot eval-hands drop ] if ; recursive

: total-wins ( -- x ) 0 num-wins keys [ num-wins at + ] each ;

: output-wins ( -- ) num-wins keys [ dup [ print-card ] each num-wins at total-wins / 100 * >float . ] each ;

: calculate-probability ( com hands discard -- ) deck remove-used-cards extend output-wins clear-num-wins ;
